
/**************************************************************************
 *
 *  $Id: mbgstatus.c 1.14 2017/07/05 18:34:46 martin REL_M $
 *
 *  Description:
 *    Main file for mbgstatus program which demonstrates how to
 *    access a Meinberg device via IOCTL calls and prints device
 *    information.
 *
 * -----------------------------------------------------------------------
 *  $Log: mbgstatus.c $
 *  Revision 1.14  2017/07/05 18:34:46  martin
 *  New way to maintain version information.
 *  Support build under Windows.
 *  Show many more details, at different verbosity levels.
 *  Use more functions from common library modules.
 *  Use codes and inline functions from mbgerror.h.
 *  Proper return codes and exit codes.
 *  Revision 1.13  2009/09/29 15:02:16  martin
 *  Updated version number to 3.4.0.
 *  Revision 1.12  2009/07/24 14:02:59  martin
 *  Display LAN and PTP status of PTP cards.
 *  Updated version number to 3.3.0.
 *  Revision 1.11  2009/06/19 14:20:36  martin
 *  Display raw IRIG time with TCR cards which support this.
 *  Revision 1.10  2009/06/16 08:21:08  martin
 *  Intermediate version 3.1.0a.
 *  Display IRIG debug status, if supported by the card.
 *  Revision 1.9  2009/03/20 11:35:41  martin
 *  Updated version number to 3.1.0.
 *  Updated copyright year to include 2009.
 *  Display signal source after signal level.
 *  Display GPS UTC parameter info, if supported by the card.
 *  Display IRIG control bits, if supported by the card.
 *  Revision 1.8  2008/12/22 12:48:18  martin
 *  Updated description, copyright, revision number and string.
 *  Use unified functions from toolutil module.
 *  Warn if a PCI Express device with unsafe IRQ support is detected.
 *  Account for signed irq_num.
 *  Accept device name(s) on the command line.
 *  Don't use printf() without format, which migth produce warnings
 *  with newer gcc versions.
 *  Revision 1.7  2007/07/24 09:33:52  martin
 *  Fixed display of port and IRQ resources.
 *  Updated copyright to include 2007.
 *  Revision 1.6  2006/03/10 12:38:22  martin
 *  Fixed printing of sign in print_position().
 *  Revision 1.5  2004/11/08 15:41:56  martin
 *  Modified formatted printing of date/time string.
 *  Using type casts to avoid compiler warnings.
 *  Revision 1.4  2003/07/30 08:16:39  martin
 *  Also displays oscillator DAC values for GPS.
 *  Revision 1.3  2003/07/08 15:38:57  martin
 *  Call mbg_find_devices().
 *  Account for swap_doubles() now being called inside
 *  the mbgdevio API functions.
 *  Show IRQ number assigned to a device.
 *  Revision 1.2  2003/04/25 10:28:05  martin
 *  Use new functions from mbgdevio library.
 *  New program version v2.1.
 *  Revision 1.1  2001/09/17 15:08:59  martin
 *
 **************************************************************************/

// include Meinberg headers
#include <mbgdevio.h>
#include <mbgutil.h>
#include <mbgtime.h>
#include <pcpslstr.h>
#include <toolutil.h>  // common utility functions
#include <lan_util.h>
#include <deviohlp.h>
#include <timeutil.h>
#include <str_util.h>
#include <nanotime.h>

// include system headers
#include <stdio.h>
#include <stdlib.h>


#define MBG_MICRO_VERSION          0
#define MBG_FIRST_COPYRIGHT_YEAR   2001
#define MBG_LAST_COPYRIGHT_YEAR    0      // use current year by default

static const char *pname = "mbgstatus";


static int loops;
static int must_list_device_names;
static long sleep_secs;
static long sleep_usecs;
static unsigned int verbose;

static const char *ref_name[N_PCPS_REF]= PCPS_REF_NAMES_ENG;
static const char *icode_rx_names[N_ICODE_RX] = DEFAULT_ICODE_RX_NAMES;
static const char *osc_name[N_GPS_OSC] = DEFAULT_GPS_OSC_NAMES;

static int year_limit = 1990;

static int max_ref_offs_h = MBG_REF_OFFS_MAX / MINS_PER_HOUR;
static int invt_reason;

static const char *wdays[7] = { "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat" };

LANGUAGE language;
CTRY ctry;



static /*HDR*/
void show_invt_reason( void )
{
  static const char fmt[] = "\n** Warning: %s\nThe command %s.\n";

  switch ( invt_reason )
  {
    case 2:
      printf( fmt, DEFAULT_STR_IRIG_NOT_CFGD_EN,
              "\"mbgirigcfg\" can be used to change the settings" );
      break;

    case 1:
      printf( fmt, DEFAULT_STR_IRIG_INVT_EN,
              "\"mbgctrl DATE=...\" can be used to set the on-board date" );
      break;

  }  // switch

}  // show_invt_reason



static /*HDR*/
void print_pcps_time( const char *s, const PCPS_TIME *tp, const char *tail )
{
  const char *fmt = "%s";
  char ws[256];

  if ( s )
    printf( fmt, s );

  printf( fmt, pcps_date_time_str( ws, sizeof( ws ), tp, year_limit,
          pcps_tz_name( tp, PCPS_TZ_NAME_FORCE_UTC_OFFS, 0 ) ) );

  if ( ( verbose > 0 ) && _pcps_time_is_read( tp ) )
    printf( ", st: 0x%02lX", (ulong) tp->status );

  if ( tail )
    printf(  fmt, tail );

}  // print_pcps_time



static /*HDR*/
void print_dms( const char *s, const DMS *p, const char *tail )
{
  const char *fmt = "%s";

  printf( "%s %c %3i deg %02i min %06.3f sec",
          s,
          p->prefix,
          p->deg,
          p->min,
          p->sec
        );

  if ( tail )
    printf( fmt, tail );

}  // print_dms



static /*HDR*/
void print_position( const char *s, const POS *p, const char *tail )
{
  const char *fmt = "%s";
  double r2d = 180 / PI;


  if ( s )
  {
    printf( fmt, s );

    if ( verbose )
      printf( "\n" );
  }

  if ( verbose > 1 )
  {
    printf( "  x: %.0fm y: %.0fm z: %.0fm",
            p->xyz[XP], p->xyz[YP], p->xyz[ZP] );

    if ( tail )
      printf( fmt, tail );
  }

  // LLA latitude and longitude are in radians, convert to degrees
  printf( "  lat: %+.4f lon: %+.4f alt: %.0fm",
          p->lla[LAT] * r2d, p->lla[LON] * r2d, p->lla[ALT] );

  if ( tail )
    printf( fmt, tail );

  if ( verbose )
  {
    print_dms( "  latitude: ", &p->latitude, tail );
    print_dms( "  longitude:", &p->longitude, tail );
  }


}  // print_position



static /*HDR*/
void show_signal( MBG_DEV_HANDLE dh, const PCPS_DEV *pdev, int signal )
{
  int ref_type;
  int rc;

  ref_type = _pcps_ref_type( pdev );

  if ( ref_type >= N_PCPS_REF )
    ref_type = PCPS_REF_NONE;

  printf( "Signal: %u%%  (%s", signal * 100 / PCPS_SIG_MAX, ref_name[ref_type] );

  if ( _pcps_is_irig_rx( pdev ) )
  {
    IRIG_INFO irig_rx_info;
    MBG_REF_OFFS ref_offs;

    rc = mbg_get_irig_rx_info( dh, &irig_rx_info );

    if ( rc == MBG_SUCCESS )
    {
      int idx = irig_rx_info.settings.icode;

      if ( idx < N_ICODE_RX )
      {
        printf( " %s", icode_rx_names[idx] );

        if ( !( MSK_ICODE_RX_HAS_TZI & ( 1UL << idx ) ) )
        {
          if ( _pcps_has_ref_offs( pdev ) )
          {
            rc = mbg_get_ref_offs( dh, &ref_offs );

            if ( rc == MBG_SUCCESS )
            {
              int ref_offs_h = ref_offs / MINS_PER_HOUR;

              if ( abs( ref_offs_h ) > max_ref_offs_h )
                printf( ", ** UTC offs not configured **" );
              else
                printf( ", UTC%+ih", ref_offs_h );
            }
          }
        }
      }
    }
  }
  else
    if ( _pcps_has_pzf( pdev ) )
      printf( "/PZF" );

  printf( ")\n" );

}  // show_signal



static /*HDR*/
void show_time_and_status( MBG_DEV_HANDLE dh, const PCPS_DEV *pdev, const char *tail )
{
  const char status_fmt[] = "Status info: %s%s\n";
  const char status_err[] = "*** ";
  const char status_ok[] = "";
  const char *info_err = ( _pcps_is_gps( pdev ) || _pcps_is_lwr( pdev ) ) ?
                         "ANTENNA FAULTY" : "NO INPUT SIGNAL";
  const char info_ok[] = "Input signal available";
  PCPS_TIME t;
  PCPS_STATUS_STRS strs;
  int signal;
  int i;
  int rc = mbg_get_time( dh, &t );

  if ( mbg_cond_err_msg( rc, "mbg_get_time" ) )
    return;


  print_pcps_time( "Date/time:  ", &t, tail );

  if ( ( verbose > 0 ) && _pcps_has_hr_time( pdev ) )
  {
    PCPS_HR_TIME ht;
    char ws[80];

    rc = mbg_get_hr_time( dh, &ht );

    if ( mbg_cond_err_msg( rc, "mbg_get_hr_time" ) )
      return;

    mbg_snprint_hr_time( ws, sizeof( ws ), &ht, 0 );  // raw timestamp?
    printf( "Local HR time:  %s", ws );

    if ( verbose > 0 )
      printf( ", st: 0x%04lX", (ulong) ht.status );

    printf( "%s", tail );
  }

  signal = t.signal - PCPS_SIG_BIAS;

  if ( signal < 0 )
    signal = 0;
  else
    if ( signal > PCPS_SIG_MAX )
      signal = PCPS_SIG_MAX;

  if ( _pcps_has_signal( pdev ) )
    show_signal( dh, pdev, signal );

  if ( _pcps_has_pzf( pdev ) )
  {
    mbg_show_pzf_corr_info( dh, 0 );
    printf( "\n" );
  }

  if ( verbose && _pcps_has_irig_time( pdev ) )
  {
    PCPS_IRIG_TIME it;

    rc = mbg_get_irig_time( dh, &it );

    if ( !mbg_cond_err_msg( rc, "mbg_get_irig_time" ) )
      printf( "Raw IRIG time: yday %u, %02u:%02u:%02u\n",
              it.yday, it.hour, it.min, it.sec );
  }

  printf( status_fmt,
          ( signal < PCPS_SIG_ERR ) ? status_err : status_ok,
          ( signal < PCPS_SIG_ERR ) ? info_err : info_ok );

  // Evaluate the status code and setup status messages.
  pcps_status_strs( t.status, _pcps_time_is_read( &t ),
                    _pcps_is_gps( pdev ), &strs );

  // Print the status messages.
  for ( i = 0; i < N_PCPS_STATUS_STR; i++ )
  {
    PCPS_STATUS_STR *pstr = &strs.s[i];
    if ( pstr->cp )
      printf( status_fmt,
              pstr->is_err ? status_err : status_ok,
              pstr->cp );
  }

  invt_reason = 0;

  if ( _pcps_is_irig_rx( pdev ) && ( t.status & PCPS_INVT ) )
  {
    MBG_REF_OFFS ref_offs;

    rc = mbg_get_ref_offs( dh, &ref_offs );

    if ( !mbg_cond_err_msg( rc, "mbg_get_ref_offs" ) )
    {
      if ( _pcps_ref_offs_out_of_range( ref_offs ) )
        invt_reason = 2;
      else
        invt_reason = 1;
    }
  }

}  // show_time_and_status



static /*HDR*/
void show_sync_time( MBG_DEV_HANDLE dh, const char *tail )
{
  PCPS_TIME t;
  int rc = mbg_get_sync_time( dh, &t );

  if ( mbg_cond_err_msg( rc, "mbg_get_sync_time" ) )
    return;

  print_pcps_time( "Last sync:  ", &t, tail );

}  // show_sync_time



static /*HDR*/
void show_ext_stat_info( MBG_DEV_HANDLE dh, const PCPS_DEV *p_dev, const char *tail )
{
  const char *fmt = "%s";
  RECEIVER_INFO ri = { 0 };
  ALL_GNSS_INFO agi = { { 0 } };
  GNSS_SAT_INFO *p_gsi;
  char ws[80];
  const char *cp;
  int i;
  int rc;

  // first collect all information

  rc = mbg_setup_receiver_info( dh, p_dev, &ri );

  if ( mbg_cond_err_msg( rc, "mbg_setup_receiver_info" ) )
    return;

  if ( mbg_rc_is_success( mbg_chk_dev_is_gps( dh ) ) )
  {
    rc = mbg_chk_get_all_gnss_info( dh, &agi );

    if ( mbg_cond_err_msg( rc, "mbg_chk_get_gnss_info" ) )
      return;
  }

  // now print information

    if ( verbose )
      printf( "Feature mask: 0x%08lX\n", (ulong) ri.features );

  if ( _pcps_has_stat_info( p_dev ) )
  {
    if ( _pcps_has_stat_info_mode( p_dev ) )
    {
      switch ( agi.stat_info.mode )
      {
        case AUTO_166: cp = "Normal Operation";  break;
        case WARM_166: cp = "Warm Boot";         break;
        case COLD_166: cp = "Cold Boot";         break;

        default:  // This should never happen!
          snprintf_safe( ws, sizeof( ws ), "%s mode of operation: %02Xh",
                         str_unknown, agi.stat_info.mode );
          cp = ws;

      }  // switch

      printf( "%s", cp );
    }
  }

  if ( agi.n_gnss_supp )
  {
    #if defined( DEBUG )  // TODO
      int must_print_sv_list = _pcps_has_stat_info_svs( p_dev ) && verbose;
    #else
      int must_print_sv_list = _pcps_is_gnss( p_dev ) && verbose;
    #endif
    int print_multi_lines = ( agi.n_gnss_supp > 1 ) || must_print_sv_list;

    // print multiple lines
    // otherwise append to line
    printf( print_multi_lines ? ":\n" : ", " );

    for ( i = 0; i < agi.n_gnss_supp; i++ )
    {
      static const char * const gnss_type_names[N_GNSS_TYPES] = GNSS_TYPE_STRS;
      int gnss_type;

      p_gsi = &agi.gnss_sat_info_idx[i].gnss_sat_info;
      gnss_type = p_gsi->gnss_type;

      if ( gnss_type >= N_GNSS_TYPES )
      {
        mbg_snprintf( ws, sizeof( ws ), "(%s GNSS type %i): ", str_unknown, gnss_type );
        cp = ws;
      }
      else
        cp = gnss_type_names[gnss_type];

      if ( print_multi_lines )
        printf( "  " );

      if ( agi.gnss_mode_info.settings.gnss_set & ( 1UL << gnss_type ) )
      {
        printf( "%i %s sats tracked, %i expected to be visible\n", p_gsi->good_svs, cp, p_gsi->svs_in_view );

        if ( must_print_sv_list )
        {
          int sat_idx;

          for ( sat_idx = 0; sat_idx < MAX_USED_SATS; sat_idx++ )
          {
            int sat_num = p_gsi->svs[sat_idx];

            #if !defined( DEBUG )
              if ( sat_num == 0 )
                break;
            #endif

            printf( "%s%i", ( sat_idx == 0 ) ? "  Sats: " : ", ", sat_num );
          }

          if ( sat_idx )  // the satellite list has been printed
            printf( "\n" );
        }
      }
      else
        printf( "%s reception disabled by configuration\n", cp );
    }
  }

  if ( verbose )
  {
    #if 0 //##+++++++++++++++++++++++++
      printf( "Feature mask: 0x%08lX\n", (ulong) ri.features );
    #endif

    printf( "Osc type: %s", osc_name[( ri.osc_type < N_GPS_OSC ) ? ri.osc_type : GPS_OSC_UNKNOWN] );

    if ( _pcps_has_stat_info( p_dev ) )
    {
      printf( ", DAC cal: %+i, fine: %+i",
              (int) ( agi.stat_info.dac_cal - OSC_DAC_BIAS ),
              (int) ( agi.stat_info.dac_val - OSC_DAC_BIAS ) );
    }

    puts( "" );
  }

  if ( tail )
    printf( fmt, tail );

}  // show_ext_stat_info



static /*HDR*/
void show_gps_pos( MBG_DEV_HANDLE dh, const char *tail )
{
  POS pos;
  int rc = mbg_get_gps_pos( dh, &pos );

  if ( mbg_cond_err_msg( rc, "mbg_get_gps_pos" ) )
    return;

  print_position( "Receiver Position:", &pos, tail );

}  // show_gps_pos



static /*HDR*/
void show_utc_info( MBG_DEV_HANDLE dh, const PCPS_DEV *p_dev )
{
  UTC utc;

  int rc = mbg_get_utc_parm( dh, &utc );

  if ( mbg_cond_err_msg( rc, "mbg_get_utc_parm" ) )
    return;


  if ( utc.valid )
  {
    char tmp_str[80];
    struct tm tm = { 0 };

    long t_ls_long = (long) utc.WNlsf * SECS_PER_WEEK
                   + (long) utc.DNt * SECS_PER_DAY
                   + GPS_SEC_BIAS - 1;
    time_t t_ls = cvt_to_time_t( t_ls_long );

    rc = mbg_gmtime( &tm, &t_ls );

    if ( mbg_rc_is_success( rc ) )
      mbg_snprintf( tmp_str, sizeof( tmp_str ), "at UTC midnight at the end of %s, %04i-%02i-%02i",
                    wdays[tm.tm_wday], tm.tm_year + 1900, tm.tm_mon + 1, tm.tm_mday );
    else
      snprint_gmtime_error( tmp_str, sizeof( tmp_str ), rc, t_ls, __func__  );

    if ( verbose > 1 )
    {
      //##++++++++++ utc.delta_tls = utc.delta_tlsf - 1;

      printf( "CSUM: %04X, valid: %04X\n", utc.csum, utc.valid );
      printf( "t0t: %u|%u.%07u, A0: %g A1: %g\n",
              utc.t0t.wn, utc.t0t.sec, utc.t0t.tick,
              utc.A0, utc.A1 );
      printf( "WNlsf: %u, DN: %u, offs: %i/%i\n",
              utc.WNlsf, utc.DNt, utc.delta_tls, utc.delta_tlsf );
    }

    if ( utc.delta_tls != utc.delta_tlsf )
    {
      // a leap second is currently being announced
      printf( "UTC offset transition from %is to %is due to leap second\n"
              "%s %s.\n",
              utc.delta_tls, utc.delta_tlsf,
              ( utc.delta_tls < utc.delta_tlsf ) ? "insertion" : "deletion",
              tmp_str
            );
    }
    else
      if ( verbose )
        printf( "Leap second eventually %s\n", tmp_str );
      else
        printf( "UTC offset parameter: %is, no leap second announced.\n", utc.delta_tls );
  }
  else
    puts( "** UTC parameters not valid" );


}  // show_utc_info



static /*HDR*/
void show_irig_ctrl_bits( MBG_DEV_HANDLE dh )
{
  MBG_IRIG_CTRL_BITS irig_ctrl_bits;

  int rc = mbg_get_irig_ctrl_bits( dh, &irig_ctrl_bits );

  if ( mbg_cond_err_msg( rc, "mbg_get_irig_ctrl_bits" ) )
    return;

  printf( "IRIG control bits: %08lX (hex, LSB first)", (ulong) irig_ctrl_bits );
  printf( ", TFOM: 0x%X", _pcps_tfom_from_irig_ctrl_bits( &irig_ctrl_bits ) );
  printf( "\n" );

}  // show_irig_ctrl_bits



static /*HDR*/
char *str_raw_irig_utc_offs_hours( char *s, int max_len, const MBG_RAW_IRIG_DATA *p )
{
  int n;
  long offs = ( p->data_bytes[8] & 0x08 )
            | ( ( p->data_bytes[8] >> 2 ) & 0x04 )
            | ( ( p->data_bytes[8] >> 4 ) & 0x02 )
            | ( ( p->data_bytes[8] >> 6 ) & 0x01 );

  n = mbg_snprintf( s, max_len, "%c%li", ( p->data_bytes[8] & 0x80 ) ? '-' : '+', offs );

  if ( p->data_bytes[8] & 0x02 )
    n += mbg_snprintf( &s[n], max_len - n, "%s", ".5" );

  return s;

}  // str_raw_irig_utc_offs_hours



static /*HDR*/
void show_raw_irig_data( MBG_DEV_HANDLE dh )
{
  MBG_RAW_IRIG_DATA raw_irig_data;
  char ws[80];
  int i;

  int rc = mbg_get_raw_irig_data( dh, &raw_irig_data );

  if ( mbg_cond_err_msg( rc, "mbg_get_raw_irig_data" ) )
    return;

  printf( "Raw IRIG data:" );

  for ( i = 0; i < sizeof( raw_irig_data ); i++ )
    printf( " %02X", raw_irig_data.data_bytes[i] );

  printf( " (hex)" );
  printf( ", TFOM: 0x%X", _pcps_tfom_from_raw_irig_data( &raw_irig_data ) );
  printf( ", UTC offs: %sh", str_raw_irig_utc_offs_hours( ws, sizeof( ws ), &raw_irig_data ) );
  printf( "\n" );

}  // show_raw_irig_data



static /*HDR*/
void show_irig_debug_status( MBG_DEV_HANDLE dh )
{
  static const char *status_str[N_MBG_DEBUG_BIT] = MBG_DEBUG_STATUS_STRS;

  MBG_DEBUG_STATUS st;
  int i;
  int rc = _mbg_generic_read_var( dh, PCPS_GET_DEBUG_STATUS, st );

  if ( mbg_cond_err_msg( rc, "show_irig_debug_status" ) )
    return;

  printf( "Debug status (hex): %08lX\n", (ulong) st );

  for ( i = 0; i < N_MBG_DEBUG_BIT; i++ )
    if ( st & ( 1UL << i ) )
      printf( "  %s\n", status_str[i] );

}  // show_irig_debug_status



static /*HDR*/
void show_lan_intf_state( MBG_DEV_HANDLE dh )
{
  IP4_SETTINGS ip4_settings;
  LAN_IF_INFO lan_if_info;
  char ws[100];

  int rc = mbg_get_ip4_state( dh, &ip4_settings );

  if ( mbg_cond_err_msg( rc, "mbg_get_ip4_state" ) )
    return;

  rc = mbg_get_lan_if_info( dh, &lan_if_info );

  if ( mbg_cond_err_msg( rc, "mbg_get_lan_if_info" ) )
    return;


  printf( "On-board LAN interface settings:\n" );

  snprint_mac_addr( ws, sizeof( ws ), &lan_if_info.mac_addr );
  printf( "  MAC Address:    %s\n", ws );

  snprint_ip4_addr( ws, sizeof( ws ), &ip4_settings.ip_addr, NULL );
  printf( "  IP Address:     %s%s\n", ws, ( ip4_settings.flags & IP4_MSK_DHCP ) ?
          " (DHCP)" : "" );

  snprint_ip4_addr( ws, sizeof( ws ), &ip4_settings.netmask, NULL );
  printf( "  Net Mask:       %s\n", ws );

  snprint_ip4_addr( ws, sizeof( ws ), &ip4_settings.broad_addr, NULL );
  printf( "  Broadcast Addr: %s\n", ws );

  snprint_ip4_addr( ws, sizeof( ws ), &ip4_settings.gateway, NULL );
  printf( "  Gateway:        %s\n", ws );

  printf( "  Link detected:  %s\n", ( ip4_settings.flags & IP4_MSK_LINK ) ? "YES" : "NO" );

}  // show_lan_intf_state



static /*HDR*/
void show_ptp_state( MBG_DEV_HANDLE dh )
{
  static const char *ptp_role_strs[N_PTP_ROLES] = PTP_ROLE_STRS;
  static const char *ptp_state_strs[N_PTP_PORT_STATE] = PTP_PORT_STATE_STRS;
  static const char *ptp_nw_prot_strs[N_PTP_NW_PROT] = PTP_NW_PROT_STRS;
  static const char *ptp_delay_mech_names[N_PTP_DELAY_MECH] = PTP_DELAY_MECH_NAMES;
  static const PTP_TABLE ptp_time_source_tbl[] = PTP_TIME_SOURCE_TABLE;
  static const char *ptp_clock_accuracy_strs[] = PTP_CLOCK_ACCURACY_STRS;

  char ws[100];
  const char *cp;
  int must_show_slave_mode_info;
  int must_show_master_mode_info;
  int slave_mode_active;
  int master_mode_active;
  int any_mode_active;
  int utc_offset_valid;
  PTP_STATE ptp_state;
  PTP_CFG_INFO ptp_info;
  int tmp;

  int rc = mbg_get_ptp_state( dh, &ptp_state );

  if ( mbg_cond_err_msg( rc, "mbg_get_ptp_state" ) )
    return;

  rc = mbg_get_ptp_cfg_info( dh, &ptp_info );

  if ( mbg_cond_err_msg( rc, "mbg_get_ptp_info" ) )
    return;


  // set up some flags controlling which information to be shown

  must_show_slave_mode_info = ( verbose > 1 ) ||
        ( ( ( 1UL << ptp_info.settings.ptp_role ) & PTP_ROLE_MSK_SLAVES ) != 0 );

  must_show_master_mode_info = ( verbose > 1 ) ||
        ( ( ( 1UL << ptp_info.settings.ptp_role ) & PTP_ROLE_MSK_MASTERS ) != 0 );

  slave_mode_active = ( ptp_state.port_state == PTP_PORT_STATE_UNCALIBRATED )
                   || ( ptp_state.port_state == PTP_PORT_STATE_SLAVE );

  master_mode_active = ( ptp_state.port_state == PTP_PORT_STATE_PRE_MASTER )
                    || ( ptp_state.port_state == PTP_PORT_STATE_MASTER )
                    || ( ptp_state.port_state == PTP_PORT_STATE_PASSIVE );

  any_mode_active = slave_mode_active || master_mode_active;


  // PTP role and port state

  printf( "PTP port state:\n" );

  printf( "  Port mode:       " );

  if ( any_mode_active )
    printf( "%s", ( ptp_state.flags & PTP_FLAG_MSK_IS_UNICAST ) ?
            "Unicast " : "Multicast " );

  if ( ptp_state.port_state < N_PTP_PORT_STATE )
    printf( "%s", ptp_state_strs[ptp_state.port_state] );
  else
    printf( "%s, code %i", str_undefined, ptp_state.port_state );

  if ( !any_mode_active  || ( verbose > 1 ) )
    printf( " in %s role",
            ( ptp_info.settings.ptp_role < N_PTP_ROLES ) ?
            ptp_role_strs[ptp_info.settings.ptp_role] : str_undefined );

  printf( "\n" );


  if ( !any_mode_active || verbose )
  {
    printf( "  PTP protocol:    " );

    // If not fully synchronized then not all fields of the PTP_STATE
    // structure may have been filled, so we show configuration settings
    // in this case.

    if ( ptp_state.ptp_prot_version )
      printf( "v%i, ", ptp_state.ptp_prot_version );

    tmp = any_mode_active ? ptp_state.nw_prot : ptp_info.settings.nw_prot;
    printf( "%s", ( tmp < N_PTP_NW_PROT ) ?
            ptp_nw_prot_strs[tmp] : str_undefined );

    printf( ", %s step",
            ( ptp_state.flags & PTP_FLAG_MSK_ONE_STEP ) ? "one" : "two" );

    tmp = any_mode_active ? ptp_state.domain_number : ptp_info.settings.domain_number;
    printf( ", domain %i", tmp );

    tmp = any_mode_active ? ptp_state.delay_mech : ptp_info.settings.delay_mech;
    printf( ", delay mech. %s", ( tmp < N_PTP_DELAY_MECH ) ?
              ptp_delay_mech_names[tmp] : str_undefined );

    tmp = any_mode_active ? ptp_state.log_delay_req_intv : ptp_info.settings.delay_req_intv;
    printf( ", dly req. intv. 2^%i s", tmp );

    printf( "\n" );
  }


  if ( must_show_slave_mode_info )
  {
    cp = ( slave_mode_active || ( verbose > 1 ) ) ? ws : str_not_avail;

    snprint_octets( ws, sizeof( ws ), ptp_state.gm_id.b,
                    sizeof( ptp_state.gm_id.b ), MAC_SEP_CHAR_ALT, NULL );
    printf( "  Grandmaster ID:  %s\n", cp );

    snprint_nano_time( ws, sizeof( ws ), &ptp_state.offset );
    printf( "  PTP time offset: %s\n", cp );

    snprint_nano_time( ws, sizeof( ws ), &ptp_state.path_delay );
    printf( "  PTP path delay:  %s\n", cp );

    if ( verbose > 1 )
    {
      snprint_nano_time( ws, sizeof( ws ), &ptp_state.mean_path_delay );
      printf( "  Mean path delay: %s\n", cp );

      snprint_nano_time( ws, sizeof( ws ), &ptp_state.delay_asymmetry );
      printf( "  Delay asymmetry: %s\n", cp );
    }
  }


  // PTP time scale

  cp = NULL;

  if ( ptp_state.flags & PTP_FLAG_MSK_TIMESCALE_IS_PTP )  // this is the default
  {
    if ( must_show_master_mode_info || verbose )
      cp = "TAI (standard)";
  }
  else  // unusual case, print info if available
    if ( any_mode_active )
      cp = "arbitrary (non-standard)";

  if ( cp )
    printf( "  PTP time scale:  %s\n", cp );


  // UTC offset and leap second status

  utc_offset_valid = ( ptp_state.flags & PTP_FLAG_MSK_UTC_VALID ) != 0;

  printf( "  PTP UTC offset: " );

  if ( !utc_offset_valid )
    printf( " %s", str_unknown );  // UTC offset not valid

  if ( utc_offset_valid || ( verbose > 1 ) )
  {
    printf( utc_offset_valid ? " " : " (" );

    printf( "%+i s", ptp_state.utc_offset );

    if ( ptp_state.flags & PTP_FLAG_MSK_LS_ANN )  // leap second is announced
    {
      // distinguish between leap second insertion and deletion
      printf( ", leap second %s scheduled",
              ( ptp_state.flags & PTP_FLAG_MSK_LS_ANN_NEG ) ? "deletion" : "insertion" );
    }

    if ( !utc_offset_valid )
      printf( ")" );
  }

  printf( "\n" );


  if ( verbose > 1 )
  {
    const PTP_TABLE *p_tbl;

    printf( "PTP clock state:\n" );

    for ( p_tbl = ptp_time_source_tbl; p_tbl->name; p_tbl++ )
      if ( p_tbl->value == ptp_state.time_source )
        break;

    printf( "  Time source: %s\n", p_tbl->name ? p_tbl->name : str_unknown );

    printf( "  Clock class: %i\n", ptp_state.clock_class );

    tmp = N_PTP_CLOCK_ACCURACY - PTP_CLOCK_ACCURACY_NUM_BIAS;

    if ( ( ptp_state.clock_accuracy >= PTP_CLOCK_ACCURACY_NUM_BIAS ) &&
         ( ptp_state.clock_accuracy < N_PTP_CLOCK_ACCURACY ) )
      cp = ptp_clock_accuracy_strs[ptp_state.clock_accuracy - PTP_CLOCK_ACCURACY_NUM_BIAS];
    else
      cp = str_undefined;

    printf( "  Clock accuracy: %s\n", cp );

#if 0
    time_source
    clock_class
    clock_accuracy
#endif
    printf( "  Offs sc. log var: %u", ptp_state.clock_offset_scaled_log_variance );
    printf( "\n" );
  }



}  // show_ptp_state



static /*HDR*/
int check_irq_unsafe( MBG_DEV_HANDLE dh, const PCPS_DEV *p_dev )
{
  PCPS_IRQ_STAT_INFO irq_stat_info;
  int ret_val = 0;
  int rc = mbg_get_irq_stat_info( dh, &irq_stat_info );

  if ( mbg_cond_err_msg( rc, "mbg_get_irq_stat_info" ) )
    return rc;

  if ( irq_stat_info & PCPS_IRQ_STAT_UNSAFE )
  {
    static const char *warn_line = "************************************************************************************";

    puts( "" );
    puts( warn_line );

    printf(
      "**  WARNING!\n"
      "**\n"
      "**  Device %s with S/N %s has a firmware version and ASIC version\n"
      "**  which do not allow safe operation with hardware interrupts (IRQs) enabled.\n"
      "**\n"
      "**  Please see http://www.meinberg.de/english/info/pex-upgrades.htm\n"
      "**  for information how the card can easily be upgraded, or contact\n"
      "**  Meinberg support (Email: support@meinberg.de) or your local\n"
      "**  representative.\n"
      ,
      _pcps_type_name( p_dev ), _pcps_sernum( p_dev )
    );

    if ( irq_stat_info & PCPS_IRQ_STAT_ENABLED )
    {
      printf(
        "**\n"
        "**  Interrupts are currently enabled for this card (NTP daemon running?)\n"
        "**  so other access is inhibited to prevent the system from hanging.\n"
      );

      ret_val = -1;
    }

    puts( warn_line );
    puts( "" );
  }

  return ret_val;

}  // check_irq_unsafe



static /*HDR*/
int do_mbgstatus( MBG_DEV_HANDLE dh, const PCPS_DEV *p_dev )
{
  int ret_val = 0;

  if ( check_irq_unsafe( dh, p_dev ) < 0 )
    goto done;

  for (;;)
  {
    if ( _pcps_has_gps_data( p_dev ) )
      show_ext_stat_info( dh, p_dev, NULL );

    show_time_and_status( dh, p_dev, "\n" );
    show_sync_time( dh, "\n" );

    if ( _pcps_is_gps( p_dev ) )
      show_gps_pos( dh, "\n" );

    if ( _pcps_has_utc_parm( p_dev ) && ( _pcps_is_gps( p_dev ) || ( verbose > 0 ) ) )
      show_utc_info( dh, p_dev );

    if ( verbose && _pcps_has_irig_ctrl_bits( p_dev ) )
      show_irig_ctrl_bits( dh );

    if ( verbose && ( mbg_chk_dev_has_raw_irig_data( dh ) == MBG_SUCCESS ) )
      show_raw_irig_data( dh );

    if ( verbose && _pcps_is_irig_rx( p_dev ) )
      show_irig_debug_status( dh );

    if ( mbg_chk_dev_has_lan_intf( dh ) == MBG_SUCCESS )
      show_lan_intf_state( dh );

    if ( _pcps_has_ptp( p_dev ) )
      show_ptp_state( dh );

    if ( loops > 0 )
      loops--;

    if ( loops == 0 )
      break;

    if ( sleep_secs )
      sleep( sleep_secs );
    else
      if ( sleep_usecs )
        usleep( sleep_usecs );

    printf( "\n" );

    // if this_loops is < 0 then loop forever
  }

  show_invt_reason();

done:
  return ret_val;

}  // do_mbgstatus

static MBG_DEV_HANDLER_FNC do_mbgstatus;



static /*HDR*/
void usage( void )
{
  mbg_print_usage_intro( pname,
    "This program prints status information for a device.\n"
    "The displayed information depends on the type of the card."
  );
  mbg_print_help_options();
  mbg_print_opt_info( "-c", "run continuously" );
  mbg_print_opt_info( "-l", "list device names" );
  mbg_print_opt_info( "-n num", "run num loops" );
  mbg_print_opt_info( "-s num", "sleep num seconds between calls (implies -c)" );
  mbg_print_opt_info( "-u num", "sleep num microseconds between calls (implies -c)" );
  mbg_print_opt_info( "-v", "increase verbosity" );
  mbg_print_device_options();
  puts( "" );

}  // usage



int main( int argc, char *argv[] )
{
  int c;
  int rc;

  ctry_setup( 0 );
  language = LNG_ENGLISH;
  ctry.dt_fmt = DT_FMT_YYYYMMDD;
  ctry.dt_sep = '-';

  mbg_print_program_info( pname, MBG_MICRO_VERSION, MBG_FIRST_COPYRIGHT_YEAR, MBG_LAST_COPYRIGHT_YEAR );

  // check command line parameters
  while ( ( c = getopt( argc, argv, "cln:s:u:vh?" ) ) != -1 )
  {
    switch ( c )
    {
      case 'c':
        loops = -1;
        break;

      case 'l':
        must_list_device_names = 1;
        break;

      case 'n':
        loops = atoi( optarg );
        break;

      case 's':
        sleep_secs = atoi( optarg );
        loops = -1;
        break;

      case 'u':
        sleep_usecs = atoi( optarg );
        loops = -1;
        break;

      case 'v':
        verbose++;
        break;

      case 'h':
      case '?':
      default:
        must_print_usage = 1;
    }
  }

  if ( must_print_usage )
  {
    usage();
    return MBG_EXIT_CODE_USAGE;
  }


  if ( must_list_device_names )
  {
    MBG_DEV_NAME_LIST_ENTRY *list_head = NULL;
    int n_dev = mbg_find_devices_with_names( &list_head, N_SUPP_DEV_BUS );

    if ( n_dev )
    {
      MBG_DEV_NAME_LIST_ENTRY *pos;

      printf( "Unique names of devices found:\n" );

      for ( pos = list_head; pos; pos = pos->next )
        printf( "  %s\n", pos->dev_name );
    }
    else
      printf( "No device found.\n" );

    mbg_free_device_name_list( list_head );
    printf( "\n" );

    return MBG_EXIT_CODE_SUCCESS;
  }

  if ( verbose )
    pcps_date_time_dist = 1;

  // Handle each of the specified devices.
  rc = mbg_handle_devices( argc, argv, optind, do_mbgstatus, CHK_DEV_ALL_DEVICES );

  return mbg_rc_is_success( rc ) ? MBG_EXIT_CODE_SUCCESS : MBG_EXIT_CODE_FAIL;
}
